from lemon_tcg.entities.game_state import GameState
from lemon_tcg.entities.player import Player
from lemon_tcg.libraries.game_state_library import GameStateLibrary

GAME_STATES = GameStateLibrary.get_instance()

class Game():
    def __init__(self, state: GameState) -> None:
        self.state = state

    @staticmethod
    def load_from_game_id(game_id: str) -> 'Game':
        game_state = GAME_STATES.get_by_id(id=game_id)
        if not isinstance(game_state, GameState):
            raise RuntimeError(f"Invalid game id '{game_id}'.")
        return Game(state=game_state)
    
    @staticmethod
    def new(players: list[Player]) -> 'Game':
        state = GameState.create_new(players=players)
        state.save_state()
        return Game(state=state)
    
    def save(self) -> None:
        self.state.save_state()


class EventError(Exception):
    """Exception base class for event-related errors."""

class EventNotRegisteredError(EventError):
    """Raised when it is tried to dispatch an event that was not yet registered."""

    def __init__(self, signal: str) -> None:
        self.signal = signal
        super().__init__(f"Signal '{signal}' has no designated listeners.")

class EventExecutionError(EventError):
    """Raised when an error occured while trying to execute an event listener."""

    def __init__(self, error: Exception, signal: str) -> None:
        self.error = error
        self.signal = signal
        super().__init__(f"An error occured while dispatching to signal '{signal}': {error}")


from lemon_tcg.entities.coordinates import Coordinates

BOARD_WIDTH = 5
BOARD_HEIGHT = 6

def test_user_to_global():
    TEST_VALUES = [
        (Coordinates(x=0, y=0), [Coordinates(x=0, y=3), Coordinates(x=4, y=2)]),
        (Coordinates(x=1, y=0), [Coordinates(x=1, y=3), Coordinates(x=3, y=2)]),
        (Coordinates(x=0, y=1), [Coordinates(x=0, y=4), Coordinates(x=4, y=1)]),
        (Coordinates(x=1, y=1), [Coordinates(x=1, y=4), Coordinates(x=3, y=1)]),
        (Coordinates(x=2, y=2), [Coordinates(x=2, y=5), Coordinates(x=2, y=0)]),
        (Coordinates(x=3, y=1), [Coordinates(x=3, y=4), Coordinates(x=1, y=1)]),
        (Coordinates(x=4, y=2), [Coordinates(x=4, y=5), Coordinates(x=0, y=0)])
    ]

    for original, results in TEST_VALUES:
        assert original.user_to_global(user_index=0, board_height=BOARD_HEIGHT, board_width=BOARD_WIDTH) == results[0]
        assert original.user_to_global(user_index=1, board_height=BOARD_HEIGHT, board_width=BOARD_WIDTH) == results[1]

def test_global_to_user():
    TEST_VALUES_0 = [
        (Coordinates(x=0, y=3), Coordinates(x=0, y=0)),
        (Coordinates(x=1, y=3), Coordinates(x=1, y=0)),
        (Coordinates(x=0, y=4), Coordinates(x=0, y=1)),
        (Coordinates(x=1, y=4), Coordinates(x=1, y=1)),
        (Coordinates(x=2, y=5), Coordinates(x=2, y=2)),
        (Coordinates(x=3, y=4), Coordinates(x=3, y=1)),
        (Coordinates(x=4, y=5), Coordinates(x=4, y=2))
    ]

    TEST_VALUES_1 = [
        (Coordinates(x=4, y=2), Coordinates(x=0, y=0)),
        (Coordinates(x=3, y=2), Coordinates(x=1, y=0)),
        (Coordinates(x=4, y=1), Coordinates(x=0, y=1)),
        (Coordinates(x=3, y=1), Coordinates(x=1, y=1)),
        (Coordinates(x=2, y=0), Coordinates(x=2, y=2)),
        (Coordinates(x=1, y=1), Coordinates(x=3, y=1)),
        (Coordinates(x=0, y=0), Coordinates(x=4, y=2))
    ]

    for original, result in TEST_VALUES_0:
        assert original.global_to_user(user_index=0, board_height=BOARD_HEIGHT, board_width=BOARD_WIDTH) == result

    for original, result in TEST_VALUES_1:
        assert original.global_to_user(user_index=1, board_height=BOARD_HEIGHT, board_width=BOARD_WIDTH) == result

def test_global_to_tile_index():
    TEST_VALUES = [
        (0,0), (1,0), (2,0), (3,0), (4,0),
        (0,1), (1,1), (2,1), (3,1), (4,1),
        (0,2), (1,2), (2,2), (3,2), (4,2),
        (0,3), (1,3), (2,3), (3,3), (4,3),
        (0,4), (1,4), (2,4), (3,4), (4,4),
        (0,5), (1,5), (2,5), (3,5), (4,5)
    ]

    for i, xy in enumerate(TEST_VALUES):
        x, y = xy
        assert Coordinates(x=x, y=y).global_to_tile_index(board_width=BOARD_WIDTH) == i

def test_user_to_tile_index():
    TEST_VALUES = [
        (4,2), (3,2), (2,2), (1,2), (0,2),
        (4,1), (3,1), (2,1), (1,1), (0,1),
        (4,0), (3,0), (2,0), (1,0), (0,0),
        (0,0), (1,0), (2,0), (3,0), (4,0),
        (0,1), (1,1), (2,1), (3,1), (4,1),
        (0,2), (1,2), (2,2), (3,2), (4,2)
    ]

    for i, xy in enumerate(TEST_VALUES):
        x, y = xy
        user_index = 0 if i >= 15 else 1
        assert Coordinates(x=x, y=y).user_to_tile_index(user_index=user_index, board_height=BOARD_HEIGHT, board_width=BOARD_WIDTH) == i


from lemon_tcg.constants.language import Language
from lemon_tcg.entities.config import Config

class Context():
    _instance = None

    def __init__(self) -> None:
        if Context._instance is not None:
            raise RuntimeError("Tried to initialize multiple instances of Context.")
        self.config = Config.load_state()
        
    @staticmethod
    def get_instance() -> 'Context':
        if Context._instance is None:
            Context._instance = Context()
        return Context._instance
    
    @property
    def language(self) -> Language:
        return self.config.language
    
    def clear_config(self) -> None:
        self.config = Config()

    def reset_config(self) -> None:
        self.clear_config()
        self.config.save_state()


from pytest import fixture
from lemon_tcg.context import Context

CONTEXT = Context.get_instance()

@fixture(scope="function", autouse=True)
def setup_tests():
    CONTEXT.reset_config()

    yield

    CONTEXT.reset_config()


from pydantic import BaseModel, computed_field
from lemon_tcg.event.type import EventType
from lemon_tcg.entities.coordinates import Coordinates
from lemon_tcg.entities.tile import Tile
from lemon_tcg.context import Context

CONTEXT = Context.get_instance()

class Board(BaseModel):
    height: int = CONTEXT.config.default_board_height
    width: int = CONTEXT.config.default_board_width
    tiles: list[Tile] = []
    _dispatch_id: str = ""

    @computed_field 
    @property
    def tile_count(self) -> int:
        return self.height * self.width
    
    def register_events(self, dispatch_id: str) -> None:
        self._dispatch_id = dispatch_id


from pytest import fixture
from lemon_tcg.entities.card import Card

@fixture
def babu():
    return Card.load_from_id(id="babu")

def test_load_from_id(babu: Card):
    assert isinstance(babu, Card)
    assert babu.id == "babu"


from pytest import fixture
from lemon_tcg.constants.language import Language
from lemon_tcg.entities.config import Config

@fixture
def config():
    return Config()

def test_set_language(config: Config):
    assert config.language == Language.ENGLISH
    config.set_language(language=Language.GERMAN)
    assert config.language == Language.GERMAN
    config.set_language(language=Language.FRENCH)
    assert config.language == Language.FRENCH

def test_save_state(config: Config):
    config.set_language(language=Language.GERMAN)
    config.save_state()

    new_config = Config.load_state()
    assert new_config.language == Language.GERMAN


from lemon_tcg.entities.base_save_state_entity_group import BaseSaveStateEntityGroup
from lemon_tcg.utils.file_operations import construct_path

class Deck(BaseSaveStateEntityGroup):
    FILE_PATH = construct_path("lemon_tcg/data/save/decks/{id}.json")
    name: str = "No Name"
    cards: dict[str, int] = {}


from typing import Optional
from pydantic import BaseModel
from lemon_tcg.constants.language import Language
from lemon_tcg.context import Context

CONTEXT = Context.get_instance()

class LocalizedValue(BaseModel):
    en: Optional[str] = None
    de: Optional[str] = None
    fr: Optional[str] = None

    def get_value(self, language: Optional[Language] = None) -> str:
        if isinstance(language, Language):
            language_str = language.value
        else:
            language_str = CONTEXT.language.value
        return getattr(self, language_str, None) or self.en or "missing_translation"


import os
from pathlib import Path
from typing import Optional

ROOT_DIR = str(Path(__file__).parent.parent.parent)

def construct_path(relative_path: str) -> str:
    path_parts = relative_path.split("/")
    absolute_path = os.path.join(ROOT_DIR, *path_parts)
    return absolute_path

def file_exists(file_path: str) -> bool:
    return os.path.isfile(file_path)

def file_to_string(file_path: str) -> str:
    with open(file_path, 'r', encoding='utf-8') as f:
        return f.read()
    
def string_to_file(file_path: str, data: str) -> None:
    with open(file_path, 'w', encoding='utf-8') as f:
        f.write(data)

# Returns a list of files in the given directory with a specific suffix
def files_in_directory(path: str, suffix: Optional[str] = None) -> list[str]:
    if not os.path.exists(path):
        raise ValueError(f"Directory {path} does not exist.")
    
    files = []
    for file in os.listdir(path):
        if suffix is not None:
            if suffix in file:
                files.append(file)
        else:
            files.append(file)
    return files


from lemon_tcg.constants.language import Language
from lemon_tcg.entities.base_save_state_entity import BaseSaveStateEntity
from lemon_tcg.utils.file_operations import construct_path

class Config(BaseSaveStateEntity):
    FILE_PATH = construct_path("lemon_tcg/data/save/config.json")
    language: Language = Language.ENGLISH
    default_board_height: int = 6
    default_board_width: int = 5

    def set_language(self, language: Language) -> None:
        if not isinstance(language, Language):
            raise RuntimeError(f"Tried to change config language, but given language is invalid.")
        self.language = language


from pydantic import BaseModel
from typing import ClassVar
from lemon_tcg.utils.file_operations import file_to_string, string_to_file, file_exists

class BaseSaveStateEntity(BaseModel):
    FILE_PATH: ClassVar = ""

    def save_state(self) -> None:
        data = self.model_dump_json()
        string_to_file(file_path=self.FILE_PATH, data=data)

    @classmethod
    def load_state(cls):
        if not file_exists(cls.FILE_PATH):
            data = "{}"
        else:
            data = file_to_string(file_path=cls.FILE_PATH)
        return cls.model_validate_json(json_data=data)


from pydantic import BaseModel
from typing import ClassVar
from lemon_tcg.utils.file_operations import construct_path, file_to_string, string_to_file, file_exists

class BaseSaveStateEntityGroup(BaseModel):
    FILE_PATH: ClassVar = construct_path("lemon_tcg/{id}.json")
    id: str = "unknown"

    @classmethod
    def get_file_path(cls, id: str) -> str:
        return cls.FILE_PATH.format(id=id)

    def save_state(self) -> None:
        data = self.model_dump_json()
        string_to_file(file_path=self.get_file_path(id=self.id), data=data)

    @classmethod
    def load_state(cls, id: str = "unknown"):
        if not file_exists(cls.get_file_path(id=id)):
            data = "{}"
        else:
            data = file_to_string(file_path=cls.get_file_path(id=id))
        return cls.model_validate_json(json_data=data)


from pydantic import BaseModel
from lemon_tcg.entities.localized_value import LocalizedValue
from lemon_tcg.utils.file_operations import construct_path, file_exists, file_to_string

CARD_FILE_PATH = construct_path("lemon_tcg/data/cards/{id}.json")

# A card and all its important information
class Card(BaseModel):
    id: str
    name: LocalizedValue
    description: LocalizedValue

    @staticmethod
    def load_from_id(id: str) -> 'Card':
        id = id.lower()
        file_path = CARD_FILE_PATH.format(id=id)
        if not file_exists(file_path=file_path):
            raise RuntimeError(f"Card with id '{id}' does not exist.")
        json_data = file_to_string(file_path=file_path)
        return Card.model_validate_json(json_data)


from typing import Any
import uuid
from lemon_tcg.event.type import EventType
from lemon_tcg.entities.base_save_state_entity_group import BaseSaveStateEntityGroup
from lemon_tcg.entities.board import Board
from lemon_tcg.entities.player import Player
from lemon_tcg.event.helper import register_event_listener
from lemon_tcg.event.response import EventResponse
from lemon_tcg.utils.file_operations import construct_path
from lemon_tcg.utils.time_operations import current_timestamp

class GameState(BaseSaveStateEntityGroup):
    FILE_PATH = construct_path("lemon_tcg/data/save/games/{id}.json")
    board: Board = Board()
    players: list[Player] = []
    event_tick: int = 0
    start_stamp: float = current_timestamp()
    last_update_stamp: float = 0

    def model_post_init(self, __context: Any) -> None:
        self.register_events()
        self.board.register_events(dispatch_id=self.id)
        for player in self.players:
            player.register_events(dispatch_id=self.id)

    @staticmethod
    def create_new(players: list[Player]) -> 'GameState':
        id = str(uuid.uuid4())
        return GameState(id=id, players=players)
    
    def register_events(self) -> None:
        register_event_listener(event_type=EventType.GAME_STATE_TICK, game_state_id=self.id, listener=self.tick_up)

    async def tick_up(self) -> EventResponse:
        self.event_tick += 1
        return EventResponse.success()


from pydantic import BaseModel
from lemon_tcg.utils.validator import validate_user_index

# Global coordinates start in the top left
# User coordinates start in the top left of their respective part of the board from their POV
# User 0 is at the bottom with regular POV
# User 1 is at the top with inverted POV
class Coordinates(BaseModel):
    x: int
    y: int

    def user_to_global(self, user_index: int, board_height: int, board_width: int) -> 'Coordinates':
        validate_user_index(user_index=user_index)
        half_height = board_height // 2
        if user_index == 0:
            x = self.x
            y = half_height + self.y
        else:
            x = board_width - self.x - 1
            y = half_height - self.y - 1
        return Coordinates(x=x, y=y)
    
    def global_to_user(self, user_index: int, board_height: int, board_width: int) -> 'Coordinates':
        validate_user_index(user_index=user_index)
        half_height = board_height // 2
        if user_index == 0:
            x = self.x
            y = self.y - half_height
        else:
            x = board_width - self.x - 1
            y = half_height - self.y - 1
        return Coordinates(x=x, y=y)
    
    def global_to_tile_index(self, board_width: int) -> int:
        return board_width * self.y + self.x
    
    def user_to_tile_index(self, user_index: int, board_height: int, board_width: int) -> int:
        global_coordinates = self.user_to_global(user_index=user_index, board_height=board_height, board_width=board_width)
        return global_coordinates.global_to_tile_index(board_width=board_width)
    



from pytest import fixture
from lemon_tcg.constants.language import Language
from lemon_tcg.entities.localized_value import LocalizedValue

@fixture
def full():
    return LocalizedValue(
        en="Cheese",
        de="Käse",
        fr="Fromage"
    )

@fixture
def empty():
    return LocalizedValue()

def test_get_value(full: LocalizedValue, empty: LocalizedValue):
    assert full.get_value(language=Language.GERMAN) == "Käse"
    assert full.get_value("bongo") == "Cheese" # type: ignore
    assert empty.get_value() == "missing_translation"


from pydantic import BaseModel

# A card that is currently played in the game, including all the important data associated with that
class GameCard(BaseModel):
    id: str
    owner_index: int
    meta_data: dict = {}


from setuptools import setup, find_packages

with open('README.md', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name="lemon_tcg",
    python_requires=">=3.9",
    version="0.1.1",
    packages=find_packages(),
    license="GNU General Public License v3.0",
    description="A library providing the game logic for the LemonTCG online multiplayer TCG game.",
    long_description=long_description,
    long_description_content_type='text/markdown',
    author="Zitronenjoghurt",
    install_requires=[
        'pydantic==2.6.4'
    ],
    extras_require={'dev': ['pytest', 'coverage', 'pytest-cov', 'twine', 'wheel']},
    url="https://github.com/Zitronenjoghurt/LemonTCG"
)


from pydantic import BaseModel

class Player(BaseModel):
    id: str
    index: int
    card_count: int
    display_name: str
    _dispatch_id: str = ""

    def register_events(self, dispatch_id: str) -> None:
        self._dispatch_id = dispatch_id